from dataclasses import dataclass
from abc import abstractmethod


@dataclass
class BaseRequest:
    path: str
    body: (dict, bytes)
    method: str
    headers: dict
    query: dict

    def __init__(self, raw):
        self._raw = raw

    @property
    @abstractmethod
    def authorizer(self) -> dict:
        return self._raw.get("requestContext", {}).get("authorizer")



from restless import Handler
from restless.util import camel_to_snake, snake_to_camel
from restless.openapi import make_spec, Formats
import json
from urllib.parse import unquote_plus
from flask import Response as FResponse
from flask import Flask, request
import os
from werkzeug.exceptions import BadRequest
from pydantic import BaseModel
from restless.parameters import PathParameter

THIS_FOLDER = os.path.dirname(__file__)


class Request:
    IN = "body"
    TYPE = dict

    @property
    def authorizer(self) -> dict:
        return dict(token=self.headers.get("Authorization", ''))

    def __init__(self, value, use_camel_case=True):
        self.path = unquote_plus(value.full_path.strip('?'))

        try:
            self.body = value.get_json(force=True)
        except BadRequest:
            self.body = value.data

        self.body = value.data
        self.method = value.method
        self.headers = value.headers
        self.query = value.args

        if use_camel_case:
            for member in ['body', 'headers', 'query']:
                setattr(self, member, camel_to_snake(getattr(self, member)))


class Response(FResponse):
    def __init__(self, body="", status_code=200, headers=None, use_camel_case=True):
        self.status_code = status_code
        self.headers = headers

        if isinstance(body, (bytes, str)):
            super().__init__(
                response=body,
                status=status_code,
                headers=headers
            )
            self.body = body
        elif isinstance(body, dict):
            super().__init__(
                response=snake_to_camel(body) if use_camel_case else json.dumps(body),
                status=status_code,
                headers=headers
            )
            self.body = self.response
        elif isinstance(body, BaseModel):
            super().__init__(
                response=json.dumps(snake_to_camel(body.dict())) if use_camel_case else json.dumps(body.dict()),
                status=status_code,
                headers=headers
            )
            self.body = self.response

        else:
            raise Exception("Unsupported")


class FlaskHandler(Handler):
    DATA_FORMAT = 'yml'

    def __init__(
            self, name, description, version, security=None, default_security=None, camel_case_interface=True,
            request_class=Request, response_class=Response
    ):
        self.security = security or []
        self.default_security = default_security or []
        self.name = name
        self.description = description
        self.version = version

        super().__init__(
            request=request_class, response=response_class, use_camel_case=camel_case_interface
        )

        self.schemes = ["http", "https"]

        app = Flask(__name__)

        @self.handle("get", "/spec/swagger.<extension>")
        def spec(extension: PathParameter.enum(Formats)) -> {200: str}:
            return make_spec(
                self.name,
                self.description,
                self.version,
                self,
                security=self.security,
                default_security=self.default_security,
                file_name=None,
                data_format=extension
            )

        @app.route('/', defaults={'path': ''})
        @app.route('/<path:path>')
        def catch_all(path):
            return self(request)

        self.app = app



import json
from datetime import datetime
from typing import Iterable
import re
from collections import namedtuple

C2S1 = re.compile('(.)([A-Z][a-z]+)')
C2S2 = re.compile('([a-z0-9])([A-Z])')
S2C = re.compile('(.)_([a-zA-Z])')


def camel_to_snake(obj):
    if isinstance(obj, (list, set)):
        return [
            camel_to_snake(i)
            if not isinstance(i, str) else i
            for i in obj
        ]
    elif isinstance(obj, dict):
        return {
            camel_to_snake(k): camel_to_snake(v)
            if not isinstance(v, str) else v
            for k, v in obj.items()
        }
    elif isinstance(obj, str):
        string = C2S1.sub(r'\1_\2', obj)
        return C2S2.sub(r'\1_\2', string).lower()

    return obj


def snake_to_camel(obj):
    if hasattr(obj, 'dict'):
        obj = obj.dict()

    if isinstance(obj, (list, set)):
        return [
            snake_to_camel(i)
            if not isinstance(i, str) else i
            for i in obj
        ]
    elif isinstance(obj, dict):
        return {
            snake_to_camel(k): snake_to_camel(v)
            if not isinstance(v, str) else v
            for k, v in obj.items()
        }
    elif isinstance(obj, str):
        return S2C.sub(lambda x: x[1] + x[2].upper(), obj)

    return obj


class UniversalEncoder(json.JSONEncoder):
    def default(self, obj):
        if hasattr(obj, 'dict'):
            return obj.dict()
        elif isinstance(obj, datetime):
            return obj.isoformat()
        elif isinstance(obj, str):
            return obj
        elif isinstance(obj, Iterable):
            return list(obj)

        return json.JSONEncoder.default(self, obj)


class FormData(dict):
    LOCATION = 'formData'
    DELIMITER = re.compile(b'(--.*?)\r\n')
    FILE_RE = re.compile('name="(.*)"; filename="(.*)"\r\nContent-Type: (.*)')
    KEY_VALUE_RE = re.compile('name="(.*)"')
    File = namedtuple('File', 'name data content_type')

    def __init__(self, payload: bytes):
        super().__init__()
        boundary = self.DELIMITER.match(payload).group(1)

        data_re = re.compile(
            boundary + b'\r\nContent-Disposition: form-data; (.*?)\r\n(?=' + boundary + b')', re.DOTALL
        )

        for part in data_re.findall(payload):
            names, data = part.split(b'\r\n\r\n', 1)
            names = names.decode()

            if 'filename' in names:
                name, filename, content_type = self.FILE_RE.search(names).groups()
                self[name] = self.File(filename, data, content_type)
            else:
                name = self.KEY_VALUE_RE.match(names).group(1)
                self[name] = data.decode()



import json
import yaml
from restless import Handler
from datetime import datetime
from collections import Hashable
from inspect import _empty
from restless.parameters import BinaryParameter
from restless.security import Security
from typing import List, Union
from restless.util import snake_to_camel
from enum import Enum


class Formats(Enum):
    yaml = 'yaml'
    json = 'json'


OPENAPI = "3.0.0"

TYPE_MAPPING = {
    str: "string",
    bool: "boolean",
    bytes: "file",
    datetime: "string",
    set: "array",
    int: "integer",
    float: "number",
    list: "array",
    dict: "object"
}


def make_security(security: List[Union[List[str], str]], spec: dict):
    for sec in security or []:
        if isinstance(sec, str):
            assert sec in spec['components']['securitySchemes'], spec['components']['securitySchemes']
        else:
            for s in sec:
                assert s in spec['components']['securitySchemes']

    return [
        {sec: []} if isinstance(sec, str) else {s: [] for s in sec}
        for sec in security
    ]


def get_handlers(handlers):
    if isinstance(handlers, dict):
        for v in handlers.values():
            for found in get_handlers(v):
                yield found

    else:
        yield handlers


def make_spec(
        title,
        description,
        version,
        api_handler: Handler,
        file_name='spec.yaml',
        servers=None,
        security: List[Security] = None,
        default_security: List[Union[List[str], str]] = None,
        data_format: Formats = None
):
    spec = {
        'openapi': OPENAPI,
        'tags': [],
        'info': {
            'title': title,
            'description': description,
            'version': version
        },
        'paths': {},
        'servers': servers or [],
        'components': {
            'securitySchemes': {
                sec.name: dict(sec) for sec in (security or [])
            },
            'schemas': {
                'Error': {
                    'properties': {
                        'error': {
                            'type': 'string'
                        },
                        'details': {
                            'type': 'object'
                        }
                    },
                    'required': [
                        'error'
                    ]
                }
            }
        }
    }

    if default_security:
        spec['security'] = make_security(default_security, spec)

    for sec in default_security or []:
        if isinstance(sec, str):
            assert sec in spec['components']['securitySchemes'], spec['components']['securitySchemes']
        else:
            for s in sec:
                assert s in spec['components']['securitySchemes']

    for handler in get_handlers(api_handler.handlers):
        if handler.path not in spec['paths']:
            spec['paths'][handler.path] = {}

        target = spec['paths'][handler.path]

        if handler.http_method not in target:
            target[handler.http_method] = {}

        target = target[handler.http_method]

        if handler.security is None:  # default = None
            # Uses the default security
            pass
        elif handler.security:
            # uses specific security
            target['security'] = make_security(handler.security, spec)
        else:  # security=False
            # unsecured
            target['security'] = []

        target['description'] = handler.method.__doc__ or handler.method.__name__
        target['responses'] = {
            code: {
                'description': description,
                'content': {
                    'application/json': {
                        'schema': {
                            '$ref': '#/components/schemas/Error'
                        }
                    }
                }
            }
            for code, description in [
                ('400', 'Bad Request'),
                ('401', 'Unauthorized'),
                ('403', 'Forbidden'),
                ('404', 'Not Found')
            ]
        }
        target['parameters'] = []

        if handler.tags:
            target['tags'] = handler.tags
        else:
            tokens = handler.path.split('/')

            if len(tokens) > 1:
                target['tags'] = [tokens[1]]
            else:
                target['tags'] = ['default']

        returns = handler.sig.return_annotation if isinstance(handler.sig.return_annotation, dict) else {}

        for code, model in returns.items():
            if isinstance(model, list):
                schema = {
                    "type": "array",
                    "items": {}
                }

                t = schema['items']
                model = model[0]
            else:
                schema = {}
                t = schema

            if isinstance(model, Hashable) and model in TYPE_MAPPING:
                t['type'] = TYPE_MAPPING[model]
            elif hasattr(model, 'schema'):
                t['$ref'] = '#/components/schemas/' + model.__name__
                spec['components']['schemas'][model.__name__] = model.schema()

                if api_handler.use_camel_case:
                    spec['components']['schemas'][model.__name__] = snake_to_camel(
                        spec['components']['schemas'][model.__name__]
                    )

            target['responses'][str(code)] = {
                'description': description,
                'content': {
                    'application/json': {
                        'schema': schema
                    }
                }
            }

        for param, model in handler.parameters.items():
            if getattr(model, "LOCATION", "body") not in ['body', 'formData']:
                param_spec = {
                    'name': snake_to_camel(param) if api_handler.use_camel_case else param,
                    'in': getattr(model, "LOCATION", "body"),
                    'required': handler.sig.parameters[param].default == _empty,
                    'description': model.__doc__ or param,
                    'schema': {
                        'type': 'string'
                    }
                }

                if getattr(model, 'ENUM', None):
                    param_spec['schema']['enum'] = model.enum_keys()

                target['parameters'].append(param_spec)
            else:
                if 'requestBody' not in target:
                    target['requestBody'] = {"content": {}}

                if hasattr(model, 'schema'):
                    spec['components']['schemas'][model.__name__] = model.schema()

                    if api_handler.use_camel_case:
                        spec['components']['schemas'][model.__name__] = snake_to_camel(
                            spec['components']['schemas'][model.__name__]
                        )

                    target['requestBody']['content']['application/json'] = {
                        'schema': {
                            '$ref': '#/components/schemas/' + model.__name__
                        }
                    }

                elif issubclass(model, BinaryParameter):
                    target['requestBody']['content']['application/octet-stream'] = {
                        'schema': {
                            'type': 'string',
                            'format': 'binary'
                        }
                    }

                elif getattr(model, "LOCATION", "body") == "formData":
                    if 'multipart/form-data' not in target['requestBody']['content']:
                        target['requestBody']['content']['multipart/form-data'] = {
                            'schema': {
                                'type': 'object',
                                'properties': {}
                            }
                        }

                    if issubclass(model, str):
                        target['requestBody']['content']['multipart/form-data']['schema']['properties'][param] = {
                            'type': 'string'
                        }

                    else:
                        target['requestBody']['content']['multipart/form-data']['schema']['properties'][param] = {
                            'type': 'string',
                            'format': 'binary'
                        }

    data_format = data_format or Formats.__getitem__(file_name.split('.')[1])

    if data_format == Formats.yaml:
        data = yaml.dump(spec)
    elif data_format == Formats.json:
        data = json.dumps(spec)
    else:
        raise Exception("Bad Format")

    if file_name:
        with open(file_name, 'w') as dst:
            dst.write(data)
    else:
        return data



from restless.util import camel_to_snake, snake_to_camel, UniversalEncoder
from restless.interfaces import BaseRequest
import json
from urllib.parse import unquote_plus
from base64 import b64decode, b64encode
from typing import Iterable


class Request(BaseRequest):
    @property
    def authorizer(self) -> dict:
        return self._raw.get("requestContext", {}).get("authorizer")

    def __init__(self, raw, use_camel_case=True):
        super().__init__(raw)
        self.path = unquote_plus(raw.get("path") or raw.get("rawPath"))

        if raw.get('isBase64Encoded'):
            self.body = b64decode(raw["body"].encode()) if raw.get("body") else None
        else:
            self.body = json.loads(raw["body"]) if raw.get("body") else None

        self.method = raw.get("httpMethod") or raw.get('requestContext', {}).get("http", {}).get("method")
        self.headers = raw.get("headers", {})
        self.query = raw.get("queryStringParameters") or {}

        if use_camel_case:
            for member in ['body', 'headers', 'query']:
                setattr(self, member, camel_to_snake(getattr(self, member)))


class Response(dict):
    def __init__(self, body="", status_code=200, headers=None, use_camel_case=True):
        super().__init__(
            statusCode=status_code,
            headers=headers or {}
        )

        if isinstance(body, bytes):
            self["isBase64Encoded"] = True
            self["body"] = b64encode(body).decode()
        elif isinstance(body, (dict, Iterable)):
            self["isBase64Encoded"] = False

            if isinstance(body, str):
                self["body"] = body
            else:
                if "Content-Type" not in self["headers"]:
                    self["headers"]["Content-Type"] = "application/json"

                if use_camel_case:
                    body_ = snake_to_camel(body)
                else:
                    body_ = body

                self["body"] = json.dumps(body_, cls=UniversalEncoder)
        else:
            raise Exception("Unsupported")



import json
from azure.functions import HttpRequest, HttpResponse
from restless.util import snake_to_camel, camel_to_snake, UniversalEncoder
import re
from urllib.parse import unquote_plus
from restless.interfaces import BaseRequest
from typing import Iterable


def to_camel(body: [str, bytes]):
    try:
        return json.dumps(snake_to_camel(json.loads(body)), cls=UniversalEncoder)
    except json.decoder.JSONDecodeError:
        return body


class Request(BaseRequest):
    BASE_PATH_RE = re.compile('https://.*/api')

    @property
    def authorizer(self) -> dict:
        return self._raw.params.get('code')

    def __init__(self, req: HttpRequest, use_camel_case=False):
        super().__init__(req)
        self.method = req.method
        self.headers = dict(**req.headers)
        self.query = dict(**req.params)
        self.path = unquote_plus(self.BASE_PATH_RE.sub('', req.url).split('?')[0])

        try:
            self.body = req.get_json()

            if use_camel_case:
                self.body = camel_to_snake(self.body)
        except ValueError:
            self.body = req.get_body()


class Response(HttpResponse):
    def __init__(self, *args, use_camel_case=False, **kwargs):
        if 'body' in kwargs:
            args = [kwargs['body']]
            del kwargs['body']
        else:
            args = list(args)

            if not args:
                args[0] = b''

        if isinstance(args[0], (dict, Iterable)) and not isinstance(args[0], (str, bytes)):
            args[0] = json.dumps(args[0], cls=UniversalEncoder)

            if 'headers' not in kwargs:
                kwargs['headers'] = {}

            kwargs['headers']["Content-Type"] = "application/json"

        if use_camel_case and kwargs.get('headers', {}).get("Content-Type") == "application/json":
            args[0] = json.dumps(snake_to_camel(json.loads(args[0])), cls=UniversalEncoder)

        if not isinstance(args[0], bytes):
            args[0] = args[0].encode('utf-8')

        super().__init__(*args, **kwargs)

    @property
    def dict(self) -> dict:
        return {
            'statusCode': self.status_code,
            'headers': dict(self.headers),
            'body': self.get_body().decode('utf-8')
        }



from enum import Enum
from functools import wraps
from typing import Callable
from inspect import signature


class Restrict:
    def __init__(self, check_function: Callable):
        self.check_function = check_function

        sig = signature(check_function)

        self.params = sig.parameters.keys()

    def __call__(self, func: Callable):
        sig = signature(func)

        assert all([p in sig.parameters for p in self.params])

        @wraps(func)
        def restricted(*args, **kwargs):
            self.check_function(**{p: kwargs[p] for p in self.params})

            return func(*args, **kwargs)

        return restricted


class Security(dict):
    pass


class BasicAuth(Security):
    name = 'Basic'

    def __init__(self):
        super().__init__(
            type='http',
            scheme='basic'
        )


class ApiKeyAuth(Security):
    class In(Enum):
        header = "header"
        query = "query"
        cookie = "cookie"

    def __init__(self, in_: In, name):
        self.name = name
        super().__init__(
            type='apiKey',
            name=name
        )

        self["in"] = in_.value



from restless.util import FormData
from pydantic import BaseModel
from enum import Enum
from typing import ClassVar


class StringParameter(str):
    ENUM: Enum = None

    @classmethod
    def enum(cls, enum: Enum) -> ClassVar:
        return type(cls.__name__, (cls,), {'ENUM': enum})

    @classmethod
    def enum_keys(cls):
        return [k.strip('_') for k in cls.ENUM.__members__]


class PathParameter(StringParameter):
    LOCATION = 'path'


class QueryParameter(StringParameter):
    LOCATION = 'query'


class HeaderParameter(StringParameter):
    LOCATION = 'header'


class FormParameter(StringParameter):
    LOCATION = 'formData'


class FormFile(FormData.File):
    LOCATION = 'formData'


class BinaryParameter(bytes):
    LOCATION = 'body'


class BodyParameter(BaseModel):
    pass


class AuthorizerParameter(dict):
    pass
